import math
import os
import time
import re
import warnings
import xml.etree.ElementTree as ET
from datetime import date as date_obj
from itertools import chain
from dateutil.parser import parse as parse_date
from inflection import underscore
import requests


from ip import CACHE_BASE
from ip.util import BaseSet, Model

# USPTO has a malconfigured SSL connection. Suppress warnings about it.
import urllib3


urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

LOOKUP_URL = "https://assignment-api.uspto.gov/patent/lookup"

NUMBER_CLEAN_RE = re.compile("[^\d]")
CACHE_DIR = CACHE_BASE / "uspto_assignment"
CACHE_DIR.mkdir(exist_ok=True)

session = requests.Session()
session.headers = {"Accept": "application/xml"}

class AssignmentManager:

    def filter(self, **kwargs):
        """Get assignments. 
        Args:
            patent: pat no to search
            application: app no to search
            assignee: assignee name to search
        """
        fields = {
            "patent": "PatentNumber",
            "application": "ApplicationNumber",
            "assignee": "OwnerName",
            "assignor": "PriorOwnerName",
            "pct_application": "PCTNumber",
            "correspondent": "CorrespondentName",
            "reel_frame": "ReelFrame"
        }
        for key, value in kwargs.items():
            field = fields[key]
            query = value
        if field in ["PatentNumber", "ApplicationNumber"]:
            query = NUMBER_CLEAN_RE.sub("", str(query))

        return AssignmentSet(field, query)

class AssignmentParser():
    def doc(self, element):
        data = self.xml_to_dict(element)
        data['image_url'] = f'http://legacy-assignments.uspto.gov/assignments/assignment-pat-{data["display_id"]}.pdf'
        return data

    def xml_to_list(self, element):
        output = list()
        for el in list(element):
            output.append(self.xml_to_pytype(el))
        if len(output) == 1:
            output = output[0]
        return output


    def xml_to_pytype(self, element):
        if element.text == "NULL":
            return None
        elif element.tag == "str":
            return element.text
        elif element.tag == "date":
            # To ensure the result is JSON-serializable, convert to isoformat
            date = parse_date(element.text).date()
            if date < parse_date("1776-01-01").date():
                return None
            return date.isoformat()
        elif element.tag == "int" or element.tag == "long":
            return int(element.text)


    def xml_to_dict(self, element):
        output = dict()
        for el in list(element):
            key = el.attrib["name"]
            key = underscore(key)
            if el.tag == "arr" or el.tag == "lst":
                value = self.xml_to_list(el)
            else:
                value = self.xml_to_pytype(el)
            output[key] = value
        return output


class AssignmentSet(BaseSet):
    parser = AssignmentParser()
    rows = 50

    def __init__(self, field, query):
        self.field = field
        self.query = query
        self.pages = dict()

    def __repr__(self):
        return f'<AssignmentSet>'

    def __getitem__(self, key):
        if type(key) == slice:
            indices = list(range(len(self)))[key.start : key.stop : key.step]
            return [self.__getitem__(index) for index in indices]
        else:
            if key < 0:
                key = len(self) - key
            page_no = math.floor(key / self.rows)
            offset = key - page_no * self.rows
            return Assignment(self._get_page(page_no)[offset])

    def __len__(self):
        if not hasattr(self, "_len"):
            self._get_page(0)
        return self._len

    def _get_page(self, page_no):
        if page_no not in self.pages:
            params = {
                "filter": self.field,
                "query": self.query,
                "rows": self.rows,
                "offset": page_no * self.rows,
            }
            filename = "-".join(
                [self.field, self.query, str(self.rows), str(page_no * self.rows), ".xml"]
            ).replace("/", "_")
            filename = CACHE_DIR / filename
            if filename.exists():
                text = open(filename).read()
            else:
                response = session.get(LOOKUP_URL, params=params, verify=False)
                text = response.text
                with open(filename, "w") as f:
                    f.write(text)

            self.pages[page_no] = self._parse_page(text)
        return self.pages[page_no]
    
    def _parse_page(self, text):
        tree = ET.fromstring(text.encode('utf-8'))
        result = tree.find('./result')
        self._len = int(result.attrib['numFound'])
        return [self.parser.doc(doc) for doc in result]

class Assignment(Model):
    objects = AssignmentManager()

    def __repr__(self):
        return f'<Assignment(id={self.id})>'


import os
from tempfile import TemporaryDirectory

import pytest

from ip.uspto_assignments import Assignment

class TestAssignment:
    def test_fetch_assignments_by_assignee(self):
        assignments = Assignment.objects.filter(assignee="US Well Services")
        assert len(assignments) >= 22
        assert assignments[5].pat_assignor_name_first == "OEHRING, JARED"
        assignments = Assignment.objects.filter(assignee="LogicBlox")
        assert len(assignments) >= 9

    def test_fetch_assignments_by_patent(self):
        assignments = Assignment.objects.filter(patent="8,789,601")
        assert len(assignments) >= 1

    def test_fetch_assignments_by_application(self):
        assignments = Assignment.objects.filter(application="14/190,982")
        assert len(assignments) >= 1

    def test_fetch_assignee_with_greater_than_500_assignments(self):
        assignments = Assignment.objects.filter(assignee="Borealis")
        assert len(assignments) >= 1268

    def test_get_assignment_image(self):
        assignments = Assignment.objects.filter(patent=6095661)
        assignment = assignments[0]
        assert (
            assignment.image_url
            == "http://legacy-assignments.uspto.gov/assignments/assignment-pat-038505-0128.pdf"
        )

    def test_slice_assignments(self):
        assignments = Assignment.objects.filter(assignee="US Well Services")
        assignment_list1 = [assignment.display_id for assignment in assignments[0:5]]
        assert len(assignment_list1) == 5

        assignment_list2 = [assignment.display_id for assignment in assignments[:5]]
        assert len(assignment_list2) == 5

        assignment_list3 = [assignment.display_id for assignment in assignments[-5:]]
        assert len(assignment_list3) == 5

    def test_iterate_assignments(self):
        assignments = Assignment.objects.filter(assignee="US Well Services")
        assignment_list = [assignment.display_id for assignment in assignments]
        assert len(assignment_list) == len(assignments)
    
    def test_get_aggregates(self):
        assignments = Assignment.objects.filter(assignee='Covar Applied')
        assert list(assignments.values_list('appl_num', flat=True)) == ['15274746', '15251940', '15251994', '15252319', '14938523', '14938467', '14938962', '14939089', '14284013', '14284013']
        #assert list(assignments.values_list('pat_num', flat=True)) == ['9908148', '9708898', '9708898']
        assert list(assignments.values_list('publ_num', flat=True)) == ['20170081931', '20170058620', '20170056928', '20170056929', '20160130917', '20160134843', '20160130928', '20160130889', '20140345940', '20140345940']
        #assert list(assignments.pct_applications) == []
        assert list(set(assignments.values_list('pat_assignee_name', flat=True))) == ['COVAR APPLIED TECHNOLOGIES, INC.']
        #assert list(assignments.aggregate('pat_assignor_name')) == False
        #assert False



import pytest

from ip.uspto_ptab import PtabTrial, PtabDocument

class TestPtab:
    def test_can_get_case(self):
        case = PtabTrial.objects.get('IPR2016-00831')
        assert case.patent_owner_name == 'Silicon Genesis Corporation'
    
    def test_can_get_set_of_cases(self):
        cases = PtabTrial.objects.filter(patent_owner_name='Silicon Genesis Corporation').order_by('trial_number')
        assert cases.count() == 3
        assert list(cases.all())[0].trial_number == 'IPR2016-00831'
        assert len(list(cases)) == 3

    def test_can_get_case_documents(self):
        case = PtabTrial.objects.get('IPR2016-00832')
        assert case.documents.count() == 77
        assert case.documents.first().title == 'U.S. Provisional Application No. 60/046,276'
        assert case.documents.values_list('title', flat=True)[:5] == ['U.S. Provisional Application No. 60/046,276', 'U.S. Patent No. 5,374,564', 'L. Wegmann, Historical Perspective and Future Trends for Ion Implantation Systems (1981)', 'U.S. Patent No. 5,559,043', 'Prosecution History of U.S. Patent No. 6,013,563']
    
    def test_can_get_case_from_document(self):
        doc = PtabDocument.objects.get(230910)
        assert doc.trial.trial_number == 'IPR2016-00831'
    
    def test_can_download_document(self, tmpdir):
        doc = PtabDocument.objects.get(230910)
        doc.download(path=tmpdir)
        expected_doc = tmpdir / 'U.S. Provisional Application No. 60_046,276.pdf'
        assert expected_doc.exists()


import requests
import json
from copy import deepcopy
from ip import CACHE_BASE
from ip.util import BaseSet, hash_dict, one_to_one, one_to_many, Model
import inflection
import mimetypes
import os
import shutil
from dateutil.parser import parse as parse_dt

CACHE_DIR = CACHE_BASE / 'ptab'
CACHE_DIR.mkdir(exist_ok=True)

session = requests.Session()

class PtabManager(BaseSet):
    page_size = 25

    def __init__(self, *args, **kwargs):
        self.args = args 
        self.kwargs = kwargs
        self.obj_class = globals().get(self.obj_class)
    
    def filter(self, *args, **kwargs):
        return self.__class__(*args, *self.args, **{**kwargs, **self.kwargs})
    
    def get(self, *args, **kwargs):
        manager = self.filter(*args, **kwargs)
        count = manager.count()
        if count > 1:
            raise ValueError('More than one result!')
        elif count == 0:
            raise ValueError('Object Not Found!')
        else:
            return manager.first()

    def __getitem__(self, key):
        if type(key) == slice:
            indices = list(range(len(self)))[key.start : key.stop : key.step]
            return [self.__getitem__(index) for index in indices]
        else:
            offset = int(key / self.page_size) * self.page_size
            position = (key % self.page_size)
            data = self.request(dict(offset=offset))
            results = data['results']
            return self.obj_class(results[position])
    
    def order_by(self, *args):
        kwargs = deepcopy(self.kwargs)
        if 'sort' not in kwargs:
            kwargs['sort'] = list()
        kwargs['sort'] += args
        return self.__class__(*self.args, **kwargs)

    def first(self):
        return self[0]

    def __len__(self):
        return self.count()

    def count(self):
        response = self.request()
        response_data = response
        count = response_data['metadata']['count']
        return count
        
    def request(self, params=dict()):
        params = {**{self.primary_key: self.args}, **self.kwargs, **params}
        params = {inflection.camelize(k, uppercase_first_letter=False): v for (k, v) in params.items()}
        fname = CACHE_DIR / f'{self.__class__}-{hash_dict(params)}.json'
        if not fname.exists():
            response = session.get(self.base_url, params=params)
            with open(fname, 'w') as f:
                json.dump(response.json(), f, indent=True)
        return json.load(open(fname))

    def all(self):
        return iter(self)
        
class PtabTrialManager(PtabManager):
    base_url = 'https://ptabdata.uspto.gov/ptab-api/trials'
    obj_class = 'PtabTrial'
    primary_key = 'trial_number'

class PtabTrial(Model):
    objects = PtabTrialManager()
    us_application = one_to_one('ip.USApplication', appl_id='application_number')
    documents = one_to_many('ip.PtabDocument', trial_number='trial_number')

    def __repr__(self):
        return f'<PtabTrial(trial_number={self.trial_number})>'
    
class PtabDocumentManager(PtabManager):
    base_url = 'https://ptabdata.uspto.gov/ptab-api/documents'
    obj_class = 'PtabDocument'
    primary_key = 'id'

class PtabDocument(Model):
    objects = PtabDocumentManager()
    trial = one_to_one('ip.PtabTrial', trial_number='trial_number')

    def download(self, path='.'):
        url = self.links[1]['href']
        extension = mimetypes.guess_extension(self.media_type)
        base_name = self.title.replace('/', '_') + extension
        cdir = os.path.join(CACHE_DIR, self.trial_number)
        os.makedirs(cdir, exist_ok=True)
        cname = os.path.join(cdir, base_name)
        oname = os.path.join(path, base_name)
        if not os.path.exists(cname):
            response = session.get(url, stream=True)
            with open(cname, 'wb') as f:
                for chunk in response.iter_content(1024):
                    f.write(chunk)
        shutil.copy(cname, oname)

    def __repr__(self):
        return f'<PtabDocument>'




import lxml.etree as ET
import re
import os
import math
import requests
from tempfile import TemporaryDirectory
from PyPDF2 import PdfFileMerger
import json
from hashlib import md5
from ip import SETTINGS, CACHE_BASE
from collections import namedtuple

from ip.util import BaseSet, one_to_many, one_to_one

# XML Namespaces for ElementTree
NS = {
    "ops": "http://ops.epo.org",
    "epo": "http://www.epo.org/exchange",
    "ft": "http://www.epo.org/fulltext",
    "reg": "http://www.epo.org/register",
}

CLIENT_SETTINGS = SETTINGS['EpoOpenPatentServices']
KEY = CLIENT_SETTINGS['ApiKey']
SECRET = CLIENT_SETTINGS['Secret']
CACHE_DIR = CACHE_BASE / 'epo'
CACHE_DIR.mkdir(exist_ok=True)

session = requests.Session()
whitespace_re = re.compile(" +")
country_re = re.compile(r"^[A-Z]{2}")
ep_case_re = re.compile(r'EP(?P<number>[\d]+)(?P<kind>[A-Z]\d)?')


DocDB = namedtuple('DocDB', ['country', 'number', 'kind', 'date'])
EpoDoc = namedtuple('EpoDoc', ['number', 'kind', 'date'])

class OPSException(Exception):
    pass

class OPSAuthenticationException(OPSException):
    pass


class OPSParser:
    """Internal xml parsing class for common tasks"""
    def docdb_number(self, el):
        raw_data = {
            "country": el.find("./epo:country", NS),
            "number": el.find("./epo:doc-number", NS),
            "kind": el.find("./epo:kind", NS),
            "date": el.find("./epo:date", NS),
        }
        for k, v in raw_data.items():
            if v is not None:
                raw_data[k] = v.text

        return DocDB(**raw_data)
    
    def epodoc_number(self, el):
        raw_data = {
            "number": el.find("./epo:doc-number", NS),
            "kind": el.find("./epo:kind", NS),
            "date": el.find("./epo:date", NS),
        }
        for k, v in raw_data.items():
            if v is not None:
                raw_data[k] = v.text

        return EpoDoc(**raw_data)

    def cpc_class(self, el):
        keys = "section, class, subclass, main-group, subgroup".split(", ")
        data = {k: el.find("./epo:" + k, NS) for k in keys}
        data = {k: v.text for (k, v) in data.items() if v is not None}
        return f"{data.get('section', '')}{data.get('class', '')}{data.get('subclass', '')} {data.get('main-group', '')}/{data.get('subgroup', '')}"


    def pct_to_docdb(self, number):
        _, country_year, number = number.split("/")
        country, year = country_year[:2], country_year[2:]
        if len(year) == 2:
            if int(year) > 50:
                year = "19" + year
            else:
                year = "20" + year

        # DocDB format changed in 2004:
        # Pre-2003 - CCyynnnnnW
        # Post-2003 - CCccyynnnnnnW
        if int(year) >= 2004:
            case_number = year + number.rjust(6, "0")
        else:
            case_number = year[2:] + number.rjust(5, "0")
        return DocDB(country, case_number, 'W', None)


    def citation(self, el):
        phase = el.attrib["cited-phase"]
        cited_by = el.attrib["cited-by"]
        office = el.attrib.get("office", "")
        pat_cite = el.find("./epo:patcit", NS)
        if pat_cite is not None:
            citation = dict(self.docdb_number(
                pat_cite.find('./epo:document-id[@document-id-type="docdb"]', NS)
            )._asdict())
        else:
            citation = el.find("./epo:nplcit/epo:text", NS).text
        category = (
            el.find("./epo:category", NS).text
            if el.find("./epo:category", NS) is not None
            else ""
        )
        relevant_claims = el.find("./epo:rel-claims", NS)
        relevant_passages = [
            e.text for e in el.findall("./epo:rel-passage/epo:passage", NS)
        ]
        return {
            "phase": phase,
            "cited_by": cited_by,
            "office": office,
            "citation": citation,
            "category": category,
            "relevant_claims": relevant_claims.text if relevant_claims is not None else "",
            "relevant_passages": relevant_passages,
        }


class OPSManager:
    """
    Base Class for clients based on the EPO Open Patent Services
    Includes INPADOC and EPO Register
    """
    parser = OPSParser()

    def authenticate(self, key=None, secret=None):
        auth_url = "https://ops.epo.org/3.2/auth/accesstoken"
        global KEY, SECRET
        if key:
            KEY=key
            SECRET=secret

        response = session.post(
            auth_url, auth=(KEY, SECRET), data={"grant_type": "client_credentials"}
        )

        if not response.ok:
            raise OPSAuthenticationException()

        access_token = response.json()["access_token"]
        session.headers["Authorization"] = "Bearer " + access_token

    def pdf_request(self, fname, url, params=dict()):
        if os.path.exists(fname):
            return
        else:
            with open(fname, 'wb') as f:
                response = self.request(url, params, stream=True)
                f.write(response.raw.read())

    
    def request(self, url, params=dict(), stream=False):
        print(url)
        retry = 0
        while retry < 3:
            response = session.get(url, params=params, stream=stream)
            if response.ok:
                return response
            elif response.status_code in (400, 403):
                self.authenticate()
            elif not response.ok:
                tree = ET.fromstring(response.text.encode("utf-8"))
                code = tree.find("./ops:code", NS).text
                message = tree.find("./ops:message", NS).text
                raise OPSException(f"{response.status_code} - {code} - {message}")
            retry += 1
        raise OPSException("Max Retries Exceeded!")



    def xml_request(self, url, params=dict()):
        param_hash = md5(json.dumps(params, sort_keys=True).encode('utf-8')).hexdigest()
        fname = os.path.join(CACHE_DIR, f"{url[37:].replace('/', '_')}{param_hash if params else ''}.xml")
        if os.path.exists(fname):
            return open(fname).read()
        response = self.request(url, params)
        text = response.text
        with open(fname, "w") as f:
            f.write(text)
        return text
        
    
    def convert_to_docdb(self, number, doc_type):
        country = country_re.search(number)
        if country:
            country = country.group(0)
            number = number[2:]
        else:
            country = "US"
            number = number

        url = f"http://ops.epo.org/3.2/rest-services/number-service/{doc_type}/original/{country}.({number})/docdb"
        text = self.xml_request(url)
        tree = ET.fromstring(text.encode("utf-8"))
        output = tree.find("./ops:standardization/ops:output", NS)

        if doc_type == "application":
            app_ref = output.find("./ops:application-reference/epo:document-id", NS)
            return self.parser.docdb_number(app_ref)
        elif doc_type == "publication":
            pub_ref = output.find("./ops:publication-reference/epo:document-id", NS)
            return self.parser.docdb_number(pub_ref)
    
    def convert_to_epodoc(self, number, doc_type):
        url = f"http://ops.epo.org/3.2/rest-services/number-service/{doc_type}/original/{number})/epodoc"
        text = self.xml_request(url)
        tree = ET.fromstring(text.encode("utf-8"))
        output = tree.find("./ops:standardization/ops:output", NS)

        if doc_type == "application":
            app_ref = output.find("./ops:application-reference/epo:document-id", NS)
            return self.parser.epodoc_number(app_ref)
        elif doc_type == "publication":
            pub_ref = output.find("./ops:publication-reference/epo:document-id", NS)
            return self.parser.epodoc_number(pub_ref)


 

class InpadocManager(OPSManager):
    """
    INPADOC

    This is a manager class. Provides access to:
    .get => Retrieve an Inpadoc record
    .search => Search based on various criteria. Similar to Espacenet Smart Search.
    """
    def get(self, number=None, doc_type="publication", doc_db=False):
        if not doc_db:
            if 'PCT' in number:
                doc_type = 'application'
                doc_db = self.parser.pct_to_docdb(number)
            else:
                doc_db = self.convert_to_docdb(number, doc_type)
        return Inpadoc(doc_type, doc_db)
    
    def search(self, query):
        return InpadocSet(InpadocSearch(query))
    
    def filter(self, query):
        return self.search(query)

    def xml_data(self, pub, data_kind):
        """
        Acceptable kinds are "biblio", "description", "claims", "images", "legal", or "family
        """
        case_number = f'{pub.country}.{pub.number}{"." + pub.kind if pub.kind else ""}'

        if data_kind not in ("biblio", "description", "claims", "images", "legal", "family"):
            raise OPSException(data_kind + " is not a valid data kind")
        if data_kind == 'legal':
            legal_url = "http://ops.epo.org/3.2/rest-services/legal/{doc_type}/docdb/{case_number}/"
            url = legal_url.format(
                case_number=case_number,
                doc_type=pub.doc_type,
            )
        elif data_kind == 'family':
            family_url = "http://ops.epo.org/3.2/rest-services/family/{doc_type}/docdb/{case_number}/"
            url = family_url.format(case_number=case_number, doc_type=pub.doc_type)
        else:
            data_url = "http://ops.epo.org/3.2/rest-services/published-data/{doc_type}/docdb/{case_number}/{data_kind}"
            url = data_url.format(
                case_number=case_number, 
                doc_type=pub.doc_type, 
                data_kind=data_kind,
            )
        text = self.xml_request(url)
        print(url)
        tree = ET.fromstring(text.encode("utf-8"))
        return tree

class Inpadoc():
    """
    INPADOC Publication

    This is a lazy-loading object that accesses INPADOC data for the associated class.
    
    The first time any data is accessed, the result is fetched and then cached on the object.

    Data includes:
        bib_data => Basic bibliogrpahic information + Abstract
        description => Full Text Specification
        claims => Claims text
        images => Images listing
        download_images() => download a PDF of the published document
        legal => Status Information

    """

    objects = InpadocManager()
    us_application = one_to_one('ip.USApplication', publication='publication')

    def __init__(self, doc_type="publication", doc_db=False):
        self.doc_type = doc_type
        self.doc_db = doc_db
        self.country = doc_db.country
        self.number = doc_db.number
        self.kind = doc_db.kind
        self.date = doc_db.date
        self.dict = self.bib_data
        for k, v in self.dict.items():
            if not hasattr(self, k):
                setattr(self, k, v)

    def __repr__(self):
        return f'<Inpadoc({self.country}{self.number}, {self.doc_type})>'

    def __iter__(self):
        return iter(self.dict.items())

    @property
    def legal(self):
        tree = self.objects.xml_data(self, 'legal')
        legal_events = tree.findall('.//ops:legal', NS)
        output=list()
        for le in legal_events:
            row = dict()
            row['description'] = le.attrib['desc']
            row['explanation'] = le.attrib['infl']
            row['code'] = le.attrib['code'].strip()
            row['date'] = le.find('./ops:L007EP', NS).text
            output.append(row)
        output = list(sorted(output, key=lambda x: x['date']))
        return output

    @property
    def bib_data(self):
        # NOTE: For EP cases with search reports, there is citation data in bib data
        # We just currently are not parsing it
        tree = self.objects.xml_data(self, "biblio")
        document = tree.find("./epo:exchange-documents/epo:exchange-document", NS)
        data = dict()
        bib_data = document.find("./epo:bibliographic-data", NS)

        title = bib_data.find("./epo:invention-title[@lang='en']", NS)
        if title == None:
            title = bib_data.find("./epo:invention-title", NS)
        if title != None:
            data["title"] = title.text.strip()
        else:
            data["title"] = ""
        pub_data = bib_data.find(
            './epo:publication-reference/epo:document-id[@document-id-type="docdb"]',
            NS,
        )
        pub_data = dict(self.objects.parser.docdb_number(pub_data)._asdict())
        data["publication"] = pub_data['country'] + pub_data['number'] + pub_data['kind']
        data["publication_date"] = pub_data.get('date', None)
        data["country"] = pub_data['country']


        app_data = bib_data.find(
            './epo:application-reference/epo:document-id[@document-id-type="docdb"]',
            NS,
        )
        app_data = dict(self.objects.parser.docdb_number(app_data)._asdict())
        data["application"] = app_data['country'] + app_data['number']
        data["filing_date"] = app_data.get('date', None)

        intl_class = [
            whitespace_re.sub("", e.text)
            for e in bib_data.findall(
                "./epo:classifications-ipcr/epo:classification-ipcr/epo:text", NS
            )
        ]
        data["intl_class"] = intl_class

        cpc_classes = bib_data.findall(
            "./epo:patent-classifications/epo:patent-classification", NS
        )
        data["cpc_class"] = [self.objects.parser.cpc_class(el) for el in cpc_classes]

        priority_apps = bib_data.findall(
            './epo:priority-claims/epo:priority-claim/epo:document-id[@document-id-type="original"]/epo:doc-number',
            NS,
        )
        data["priority_claims"] = [e.text for e in priority_apps]

        parties = bib_data.find("./epo:parties", NS)
        data["applicants"] = [
            e.text
            for e in parties.findall(
                './epo:applicants/epo:applicant[@data-format="original"]/epo:applicant-name/epo:name',
                NS,
            )
        ]
        data["inventors"] = [
            e.text
            for e in parties.findall(
                './epo:inventors/epo:inventor[@data-format="original"]/epo:inventor-name/epo:name',
                NS,
            )
        ]
        abstract = document.find('./epo:abstract[@lang="en"]', NS)
        data["abstract"] = (
            "".join(t for t in abstract.itertext()).strip()
            if abstract is not None
            else ""
        )

        refs_cited = bib_data.findall("./epo:references-cited/epo:citation", NS)
        data["references_cited"] = [self.objects.parser.citation(c) for c in refs_cited]
        
        return data


    @property
    def description(self):
        tree = self.objects.xml_data(self, "description")
        description = tree.find(
            "./ft:fulltext-documents/ft:fulltext-document/ft:description", NS
        )
        text = "\n".join(description.itertext()).strip()
        return text

    @property
    def claims(self):
        tree = self.objects.xml_data(self, "claims")
        claims = [
            "".join(e.itertext()).strip()
            for e in tree.findall(
                './ft:fulltext-documents/ft:fulltext-document/ft:claims[@lang="EN"]/ft:claim/ft:claim-text',
                NS,
            )
        ]
        return claims

    @property
    def images(self):
        tree = self.objects.xml_data(self, "images")
        images = tree.find(
            './ops:document-inquiry/ops:inquiry-result/ops:document-instance[@desc="FullDocument"]',
            NS,
        )
        data = {
            "url": "http://ops.epo.org/rest-services/" + images.attrib["link"] + ".pdf",
            "num_pages": int(images.attrib["number-of-pages"]),
            "sections": {
                e.attrib["name"]: int(e.attrib["start-page"])
                for e in images.findall("./ops:document-section", NS)
            },
        }
        return data
    
    @property
    def family(self):
        tree = self.objects.xml_data(self, 'family')
        doc_db_list = list()
        for el in tree.findall("./ops:patent-family/ops:family-member", NS):
            pub_ref = el.find(
                './epo:publication-reference/epo:document-id[@document-id-type="docdb"]',
                NS,
            )
            if pub_ref is not None:
                doc_db_list.append((self.objects.parser.docdb_number(pub_ref), 'publication'))
            else:
                app_ref = el.find(
                    './epo:application-reference/epo:document-id[@document-id-type="docdb"]',
                    NS
                )
                doc_db_list.append((self.objects.parser.docdb_number(app_ref), 'application'))
        return InpadocSet(doc_db_list) 


    def download_images(self, path="."):
        """Download each page of images, and then consolidate into a single PDF
        Args:
            path: str(base path for file)
        """
        img_data = self.images
        dirname = CACHE_DIR / f'{self.doc_type}-{self.country}{self.number}{self.kind if self.kind else ""}'
        dirname.mkdir(exist_ok=True)
        for i in range(1, img_data["num_pages"] + 1):
            fname = dirname / ("page-" + str(i).rjust(6, "0") + ".pdf")
            if not fname.exists():
                self.objects.pdf_request(fname, img_data['url'], params={"Range": i})

        pages = list(sorted(os.listdir(dirname)))
        out_file = PdfFileMerger()
        for p in pages:
            out_file.append(os.path.join(dirname, p))
        out_fname = os.path.join(
            path, self.country + self.number + ".pdf"
        )
        out_file.write(out_fname)

class InpadocSet(BaseSet):
    def __init__(self, doc_db_list):
        self.objs = doc_db_list

    def __repr__(self):
        return f'<InpadocSet(len={len(self)})>'

    def __len__(self):
        return len(self.objs)

    def __getitem__(self, key):
        if type(key) == slice:
            indices = list(range(len(self)))[key.start : key.stop : key.step]
            return [self.__getitem__(index) for index in indices]
        else:
            doc_db, doc_type = self.objs[key]            
            return Inpadoc(doc_db=doc_db, doc_type=doc_type)

    def __add__(self, other):
        return InpadocSet(self.objs + other.objs)

    def __iter__(self):
        return (Inpadoc(doc_db=doc_db, doc_type=doc_type) for (doc_db, doc_type) in self.objs)


class InpadocSearch(OPSManager):
    """ 
    INPADOC Search

    This provides an interface to the EPO's INPADOC search functionality.
    This class is effectively a generator for accesssing paginated search results.
    This is accessed through the Inpadoc interface class.
    
    Query is in the CQL format, explained in the OPS documentation appendix
    http://documents.epo.org/projects/babylon/eponet.nsf/0/F3ECDCC915C9BCD8C1258060003AA712/$File/ops_v3.2_documentation_-_version_1.3.6_en.pdf
    """
    search_url = 'http://ops.epo.org/3.2/rest-services/published-data/search'
    parser = OPSParser()
    page_size=25

    def __init__(self, query):
        self.query = dict(q=query)
        text = self.xml_request(self.search_url, self.query)
        tree = ET.fromstring(text.encode("utf-8"))
        self.length = tree.find(".//ops:biblio-search", NS).attrib['total-result-count']
        self.pages = {1: text}
    
    def __len__(self):
        """Total number of results"""
        return int(self.length)

    def __getitem__(self, key):
        """Supports indexing and slicing results"""
        if type(key) == slice:
            indices = list(range(len(self)))[key.start : key.stop : key.step]
            return [self.__getitem__(index) for index in indices]
        else:
            page_num = math.floor(key / self.page_size) + 1
            line_num = key % self.page_size
            page = self._get_page(page_num)
            return (page[line_num], 'publication')

    def _get_page(self, page_num):
        """Internal private method for handling pages of results"""
        if page_num not in self.pages:
            start = (page_num - 1) * self.page_size + 1
            end = start + self.page_size - 1 
            query = {**self.query, **{'Range': f'{start}-{end}'}}
            self.pages[page_num] = self.xml_request(self.search_url, query)
        text = self.pages[page_num]
        tree = ET.fromstring(text.encode('utf-8'))
        results = tree.find('.//ops:search-result', NS)
        return [self.parser.docdb_number(el.find('./epo:document-id', NS)) for el in results]
      

class EpoManager(OPSManager):
    """
    EPO Manager

    This is a manager class to the EPO register. Retrieves information about
    EPO patents and applications

    """
    def get(self, number=None, doc_type="publication", doc_db=False):
        epodoc = self.convert_to_epodoc(number, doc_type)
        return Epo(epodoc)

    def xml_data(self, pub, data_kind):
        """
        Acceptable kinds are "biblio", "events", "procedural-steps"
        """
        register_url = "http://ops.epo.org/3.2/rest-services/register/{doc_type}/epodoc/{case_number}/"
        if data_kind not in ("biblio", "events", "procedural-steps"):
            raise ValueError(data_kind + " is not a valid data kind")
        url = (
            register_url.format(case_number=pub.case_number, doc_type=pub.doc_type)
            + data_kind
        )
        text = self.xml_request(url)
        tree = ET.fromstring(text.encode('utf-8'))
        return tree
        
class EpoParser(OPSParser):

    def status(self, el):
        return {
                'description': el.text,
                'code': el.attrib['status-code'],
                'date': el.attrib['change-date'],
            }

    def priority_claim(self, el):
        return {
            'kind': el.attrib['kind'],
            'number': el.find('./reg:doc-number', NS).text,
            'date': el.find('./reg:date', NS).text,
        }


    def docid(self, el):
        doc_id = el.find('.//reg:document-id', NS)
        raw = {
            'country': doc_id.find('./reg:country', NS),
            'number': doc_id.find('./reg:doc-number', NS),
            'kind': doc_id.find('./reg:kind', NS),
            'date': doc_id.find('./reg:date', NS),
        }
        return {k:v.text for (k,v) in raw.items() if v is not None}

    def party(self, el):
        addr_book = el.find('./reg:addressbook', NS)
        name = addr_book.find('./reg:name', NS)
        address = addr_book.find('./reg:address', NS)
        return {
            'name': name.text,
            'address': '\n'.join([t.strip() for t in address.itertext() if t.strip()]).strip()
        }
    
    def citation(self, el):
        phase = el.attrib["cited-phase"]
        office = el.attrib.get("office", "")
        pat_cite = el.find("./reg:patcit", NS)
        if pat_cite is not None:
            citation = self.docid(
                pat_cite
            )
        else:
            citation = el.find("./reg:nplcit/reg:text", NS).text
        category = (
            el.find("./reg:category", NS).text
            if el.find("./reg:category", NS) is not None
            else ""
        )
        relevant_passages = pat_cite.find('./reg:text', NS).text
        return {
            "phase": phase,
            "office": office,
            "citation": citation,
            "category": category,
            "relevant_passages": relevant_passages
        }
    
    def step(self, el):
        date = el.find('./reg:procedural-step-date/reg:date', NS)
        return {
            'phase': el.attrib['procedure-step-phase'],
            'description': ' - '.join([e.text for e in el.findall('./reg:procedural-step-text', NS)]),
            'date': date.text.strip() if date is not None else None,
            'code': el.find('./reg:procedural-step-code', NS).text.strip(),
        }



class Epo():
    objects = EpoManager()
    parse = EpoParser()

    def __init__(self, epodoc):
        self.epodoc = epodoc
        self.number = epodoc.number
        self.kind = epodoc.kind
        if self.kind == 'A':
            self.doc_type = 'application'
        else:
            self.doc_type = 'publication'
        
        self.dict = self.bib_data
        for k, v in self.dict.items():
            setattr(self, k, v)
    
    def __repr__(self):
        return f'<Epo(number={self.number}, kind={self.kind}, doc_type={self.doc_type})>'

    def __iter__(self):
        return self.dict.items()
    
    @property
    def case_number(self):
        number = f'{self.number}{"." + self.kind if self.kind else ""}'
        return number
    
    @property
    def procedural_steps(self):
        tree = self.objects.xml_data(self, 'procedural-steps')
        doc = tree.find('.//reg:register-document', NS)
        return [self.parse.step(el) for el in doc.find('./reg:procedural-data', NS)]

    @property
    def bib_data(self):
        tree = self.objects.xml_data(self, 'biblio')
        doc = tree.find('.//reg:register-document', NS)
        bib = doc.find('./reg:bibliographic-data', NS)
        parties = bib.find('./reg:parties', NS)
        output = dict()
        output['status'] = [self.parse.status(el) for el in doc.find('./reg:ep-patent-statuses', NS)]
        output['publications'] = [self.parse.docid(el) for el in bib.findall('./reg:publication-reference', NS)]
        output['intl_class'] = [''.join(el.itertext()).strip() for el in bib.find('./reg:classifications-ipcr', NS)]
        output['applications'] = [self.parse.docid(el) for el in bib.findall('./reg:application-reference', NS)]
        output['filing_language'] = bib.find('./reg:language-of-filing', NS).text
        output['priority_claims'] = [self.parse.priority_claim(el) for el in bib.find('reg:priority-claims', NS)]
        output['applicants'] = [self.parse.party(el) for el in parties.find('./reg:applicants', NS)]
        output['inventors'] = [self.parse.party(el) for el in parties.find('./reg:inventors', NS)]
        output['designated_states'] = [c.strip() for c in bib.find('./reg:designation-of-states', NS).itertext() if c.strip()]
        output['title'] = bib.find('./reg:invention-title[@lang="en"]', NS).text
        output['citations'] = [self.parse.citation(el) for el in bib.find('./reg:references-cited', NS)]
        return output


from collections import OrderedDict
from hashlib import md5
import json
from importlib import import_module
from dateutil.parser import parse as parse_dt
import inflection

def hash_dict(dictionary):
    return md5(json.dumps(dictionary, sort_keys=True).encode('utf-8')).hexdigest()


FILTERS = {
    #'exact',
    #'iexact',
    #'contains',
    #'icontains',
    'in': lambda x, y: x in y,
    'eq': lambda x, y: x == y,
    #'gt',
    #'gte',
    #'lt',
    #'lte',
    #'startswith',
    #'istartswith',
    #'endswith',
    #'iendswith',
    #'range',
    #'date',
    #'year','
    #'month',
    #'day',
    #'week',
    #'week_day',
    #'quarter','
    #'time',
    #'hour','
    #'minute',
    #'second',
    #'isnull',
    #'regex',
    #'iregex',
}


def one_to_one(class_name, **mapping):
    module_name, class_name = class_name.rsplit('.', 1)
    @property
    def get(self):
        klass = getattr(import_module(module_name), class_name)
        filter_obj = {k: getattr(self, v) for (k, v) in mapping.items()}
        return klass.objects.get(**filter_obj)
    return get

def one_to_many(class_name, **mapping):
    module_name, class_name = class_name.rsplit('.', 1)
    @property
    def get(self):
        klass = getattr(import_module(module_name), class_name)
        filter_obj = {k: getattr(self, v) for (k, v) in mapping.items()}
        return klass.objects.filter(**filter_obj)
    return get


def recur_accessor(obj, accessor):
    if '__' not in accessor:
        a = accessor
        rest = None
    else:
        a, rest = accessor.split('__', 1)
    
    if hasattr(obj, a):
        new_obj = getattr(obj, a)
        if callable(new_obj):
            new_obj = new_obj()
    else:
        try:
            a = int(a)
        except ValueError:
            pass
        try:
            new_obj = obj[a]
        except (KeyError, TypeError):
            new_obj=None
    if not rest:
        return new_obj
    else:
        return recur_accessor(new_obj, rest)

class Model(object):
    def __init__(self, data):
        self.dict = {inflection.underscore(k):v for (k,v) in data.items()}
        for k, v in self.dict.items():
            try:
                if 'datetime' in k and type(v) == str:
                    self.dict[k] = parse_dt(v)
                elif 'date' in k and type(v) == str:
                    self.dict[k] = parse_dt(v).date()
            except ValueError: # Malformed datetimes:
                self.dict[k] = None
            setattr(self, k, self.dict[k])


class BaseSet:
    def filter(self, **kwargs):
        for key, value in kwargs.items():
            selector, operator = key.rsplit('__', 1)
            if operator not in FILTERS:
                operator = 'eq'
                accessor = key
            selector_lambda = lambda x: accessor(x, accessor)
            self.objs = filter(lambda x: FILTERS[operator](selector_lambda(x), value), self.objs)

    def values(self, *fields):
        return ValuesSet(self, *fields)
    
    def values_list(self, *fields, flat=False):
        return ValuesSet(self, *fields, values_list=True, flat=flat)
    
class ValuesSet:
    def __init__(self, objs, *fields, values_list=False, flat=False):
        self.objs = objs
        self.fields = fields
        self.values_list = values_list
        self.flat = flat

    def __len__(self):
        return len(self.objs)

    def __repr__(self):
        return f'<ValuesSet(objs={repr(self.objs)})>'

    def __getitem__(self, key):
        if type(key) == slice:
            indices = list(range(len(self)))[key.start : key.stop : key.step]
            return [self.__getitem__(index) for index in indices]
        else:
            if key < 0:
                key = len(self) - key
            obj = self.objs[key]
            data = obj.dict
            if self.fields:
                fdata = OrderedDict()
                for k in self.fields:
                    key = k.replace('__', '_')
                    value = recur_accessor(obj, k)
                    fdata[key] = value 
                data = fdata
            if self.values_list:
                data = tuple(data[k] for k, v in data.items())
                if len(self.fields) == 1 and self.flat:
                    data = data[0]
            return data


import pytest
import os
import json

from ip.epo_ops import Epo, Inpadoc

FIXTURES = os.path.join(os.path.dirname(__file__), "fixtures")


class TestInpadoc:
    def test_can_get_epo_pub(self):
        pub = Inpadoc.objects.get("CA2944968")
        bib_data = pub.bib_data
        images = pub.images
        claims = pub.claims
        description = pub.description
        assert pub.title == "WIRELINE POWER SUPPLY DURING ELECTRIC POWERED FRACTURING OPERATIONS"
        assert bib_data == {
            "title": "WIRELINE POWER SUPPLY DURING ELECTRIC POWERED FRACTURING OPERATIONS",
            "publication": "CA2944968A1",
            "application": "CA2944968",
            'country': 'CA', 'filing_date': None, 'publication_date': '20170213',
            "intl_class": ["E21B41/00AI", "E21B43/26AI", "H02K7/14AI", "H02K7/18AI"],
            "cpc_class": [],
            "priority_claims": ["62/204,842"],
            "applicants": ["US WELL SERVICES LLC"],
            "inventors": ["OEHRING, JARED, ", "HINDERLITER, BRANDON"],
            "references_cited": [],
            "abstract": "A system and method for supplying electric power to various pieces of fracturing equipment in a fracturing operation with gas powered generators. The system and method also includes switch gears, auxiliary trailers, transformers, power distribution panels, new receptacles, and cables to supply three-phase power to electric fracturing equipment. The switchgear in the power supply system is weatherproof and able to endure the wear and tear of mobilization. The novel system and method provide clean and quiet electricity to all the equipment on site.",
        }
        assert len(description) == 35306
        assert claims[:5] == [
            "1. A fracturing system comprising: a turbine generator having an electrical output; an electric motor that is in electrical communication with the electrical output; a fracturing pump that is driven by the electric motor; and a wireline system that is in electrical communication with the electrical output.",
            "2. The system of claim 1, further comprising a variable frequency drive connected to the electric motor to control the speed of the motor, wherein the variable frequency drive frequently performs electric motor diagnostics to prevent damage to the at least one electric motor.",
            "3. The system of claim 1, wherein the wireline system comprises a wireline tool that is disposable in a wellbore that is selected from the group consisting of a perforating gun, a plug, a formation logging tool, a cutting tool, a casing imaging tool, and combinations thereof.",
            "4. The system of claim 1, further comprising trailers that contain at least one power distribution panel that supplies power to the hydraulic fracturing equipment.",
            "5. The system of claim 4, where the trailers further contain receptacles for attaching cable to the hydraulic fracturing equipment and cables that can sustain the power draw of the turbines with three separate plugs for the three phase power.",
        ]
        assert images == {
            "url": "http://ops.epo.org/rest-services/published-data/images/CA/2944968/A1/fullimage.pdf",
            "num_pages": 29,
            "sections": {"ABSTRACT": 1, "CLAIMS": 22, "DESCRIPTION": 2, "DRAWINGS": 25},
        }

    def test_can_download_full_images(self, tmpdir):
        pub = Inpadoc.objects.get("CA2944968")
        pub.download_images(path=str(tmpdir))
        assert os.listdir(str(tmpdir)) == ["CA2944968.pdf"]

    def test_can_handle_russian_cases(self):
        pub = Inpadoc.objects.get("RU2015124071")
        bib_data = pub.bib_data
        assert bib_data == {
            "title": "БУРОВОЕ ДОЛОТО ДЛЯ БУРИЛЬНОГО УСТРОЙСТВА",
            "publication": "RU2015124071A",
            "application": "RU2015124071",
            'country': 'RU', 'filing_date': None, 'publication_date': '20170110',
            "intl_class": ["E21B7/04AI"],
            "cpc_class": [
                "E21B 7/064",
                "E21B 17/04",
                "E21B 47/01",
                "E21B 47/12",
                "E21B 7/067",
                "E21B 10/08",
                "E21B 10/567",
            ],
            "priority_claims": ["13/683,540", "US 2013/066560"],
            "applicants": ["САЙЕНТИФИК ДРИЛЛИНГ ИНТЕРНЭШНЛ, ИНК."],
            "inventors": ["ХЕЙСИГ Джеральд"],
            "references_cited": [],
            "abstract": "",
        }

    def test_can_get_ep_application(self):
        pub = Inpadoc.objects.get("EP13844704", doc_type="application")
        bib_data = pub.bib_data
        with open(os.path.join(FIXTURES, "epo_ep.json"), 'w') as f:
            json.dump(bib_data, f, indent=2)
        assert bib_data == json.load(open(os.path.join(FIXTURES, "epo_ep.json")))

    def test_pct(self):
        doc = Inpadoc.objects.get("PCT/US16/15853")
        print(doc)
        bib_data = doc.bib_data
        with open(os.path.join(FIXTURES, "epo_pct.json"), 'w') as f:
            json.dump(bib_data, f, indent=2)
        assert bib_data == json.load(open(os.path.join(FIXTURES, "epo_pct.json")))

    def test_can_get_us_application(self):
        pub = Inpadoc.objects.get("US15915966", doc_type="application")
        bib_data = pub.bib_data
        assert bib_data == {
            "abstract": "A heading transfer unit may be used to transfer a heading from a "
            "surface base to a MWD tool. The surface base may have a master "
            "north finder to determine a heading. The heading may be "
            "transferred to the heading transfer unit, which is in turn "
            "transferred to the MWD unit. The heading on the heading transfer "
            "unit is transferred to the MWD tool.",
            "applicants": ["Scientific Drilling International, Inc"],
            "application": "US201815915966",
            "country": "US",
            "filing_date": None,
            'publication_date': '20180913',
            "cpc_class": ["E21B 47/024", "G01C 17/00"],
            "intl_class": ["E21B47/022AI", "G01C17/00AI"],
            "inventors": ["HAWKINSON, Benjamin Curtis, ", "GLEASON, Brian Daniel"],
            "priority_claims": ["62468889"],
            "publication": "US2018258752A1",
            "references_cited": [],
            "title": "DEVICE AND METHOD FOR SURVEYING BOREHOLES OR ORIENTING DOWNHOLE "
            "ASSEMBLIES",
        }

    def test_can_get_inpadoc_family(self):
        family = Inpadoc.objects.get("EP13844704", doc_type="application").family
        cases = [".".join([m.country, m.number, m.kind]) for m in family]
        assert cases == [
            "EP.2906782.A2",
            "EP.2906782.A4",
            "CA.2887530.A1",
            "CN.104968889.A",
            "RU.2015117646.A",
            "US.2014102795.A1",
            "US.9291047.B2",
            "US.2016245070.A1",
            "US.10047600.B2",
            "WO.2014059282.A2",
            "WO.2014059282.A3",
        ]

    def test_can_get_legal_status(self):
        pub = Inpadoc.objects.get("CA2300029")
        print(pub.legal)
        assert pub.legal == [
            {
                "description": "EXAMINATION REQUEST",
                "explanation": "+",
                "code": "EEER",
                "date": "2005-02-03",
            }
        ]

    def test_can_search_inpadoc(self):
        results = Inpadoc.objects.search('pa="Scientific Drilling"')
        assert len(results) == 206
        assert results.values('title')[:10] == [{'title': 'SUB-SURFACE ELECTROMAGNETIC TELEMETRY SYSTEMS AND METHODS'},
 {'title': 'DEVICE AND METHOD FOR SURVEYING BOREHOLES OR ORIENTING DOWNHOLE '
           'ASSEMBLIES'},
 {'title': 'DEVICE AND METHOD FOR SURVEYING BOREHOLES OR ORIENTING DOWNHOLE '
           'ASSEMBLIES'},
 {'title': 'METHOD FOR IMPROVING SURVEY MEASUREMENT DENSITY ALONG A BOREHOLE'},
 {'title': 'LOGGING-WHILE-DRILLING SPECTRAL AND AZIMUTHAL GAMMA RAY APPARATUS '
           'AND METHODS'},
 {'title': 'DOWNHOLE MWD SIGNAL ENHANCEMENT, TRACKING, AND DECODING'},
 {'title': 'LOGGING-WHILE-DRILLING SPECTRAL AND AZIMUTHAL GAMMA RAY APPARATUS '
           'AND METHODS'},
 {'title': 'TUMBLE GYRO SURVEYOR'},
 {'title': 'SURFACE COIL FOR WELLBORE POSITIONING'},
 {'title': 'COHERENT MEASUREMENT METHOD FOR DOWNHOLE APPLICATIONS'}]
        #us_cases = results.filter(publication__country='US')[:5]
        #from pprint import pprint
        #pprint(us_cases)
        #assert False


class TestEpoRegister:
    def test_can_get_epo_data(self):
        pub = Epo.objects.get("EP3221665A1")
        assert pub.status == [
            {
                "description": "Request for examination was made",
                "code": "15",
                "date": "20170825",
            },
            {
                "description": "The international publication has been made",
                "code": "17",
                "date": "20170512",
            },
        ]
        assert pub.bib_data == {
            "applicants": [
                {
                    "address": "16701 Greenspoint Park Dr.\n"
                    "Suite 200\n"
                    "Houston, TX 77060\n"
                    "US",
                    "name": "Scientific Drilling International, Inc.",
                }
            ],
            "applications": [
                {"country": "EP", "date": "20151119", "number": "15860753"},
                {"country": "WO", "number": "WO2015US61659"},
            ],
            "citations": [
                {
                    "category": "Y",
                    "citation": {"country": "US", "number": "2014007646"},
                    "office": "EP",
                    "phase": "international-search-report",
                    "relevant_passages": "(RODNEY PAUL F [US], et al) [Y] 1-43 * "
                    "entire document *;",
                },
                {
                    "category": "Y",
                    "citation": {"country": "US", "number": "4987684"},
                    "office": "EP",
                    "phase": "international-search-report",
                    "relevant_passages": "(ANDREAS RONALD D [US], et al) [Y] 1-14, "
                    "33-43 * entire document *;",
                },
                {
                    "category": "Y",
                    "citation": {"country": "US", "number": "2008230273"},
                    "office": "EP",
                    "phase": "international-search-report",
                    "relevant_passages": "(BROOKS ANDREW G [US]) [Y] 2-32, 35-38, "
                    "41, 43 * entire document *;",
                },
                {
                    "category": "Y",
                    "citation": {"country": "US", "number": "2007242042"},
                    "office": "EP",
                    "phase": "international-search-report",
                    "relevant_passages": "(KELLY MICHAEL B [US]) [Y] 23, 24 * "
                    "entire document *;",
                },
                {
                    "category": "Y",
                    "citation": {"country": "US", "number": "5039944"},
                    "office": "EP",
                    "phase": "international-search-report",
                    "relevant_passages": "(KIM BORIS F [US], et al) [Y] 25, 26 * "
                    "entire document *;",
                },
                {
                    "category": "A",
                    "citation": {"country": "US", "number": "2003220743"},
                    "office": "EP",
                    "phase": "international-search-report",
                    "relevant_passages": "(VAN STEENWYK DONALD H [US], et al) [A] "
                    "1-43 * entire document *",
                },
            ],
            "designated_states": [
                "EP",
                "AL",
                "AT",
                "BE",
                "BG",
                "CH",
                "CY",
                "CZ",
                "DE",
                "DK",
                "EE",
                "ES",
                "FI",
                "FR",
                "GB",
                "GR",
                "HR",
                "HU",
                "IE",
                "IS",
                "IT",
                "LI",
                "LT",
                "LU",
                "LV",
                "MC",
                "MK",
                "MT",
                "NL",
                "NO",
                "PL",
                "PT",
                "RO",
                "RS",
                "SE",
                "SI",
                "SK",
                "SM",
                "TR",
            ],
            "filing_language": "en",
            "intl_class": ["G01C25/00, E21B47/024"],
            "inventors": [
                {
                    "address": "c/o Scientific Drilling International Inc.\n"
                    "16701 Greenspoint Park Dr.\n"
                    "Suite 200\n"
                    "Houston, TX 77060\n"
                    "US",
                    "name": "VAN STEENWYK, Brett",
                }
            ],
            "priority_claims": [
                {"date": "20141119", "kind": "national", "number": "201462081944P"}
            ],
            "publications": [
                {
                    "country": "WO",
                    "date": "20160526",
                    "kind": "A1",
                    "number": "2016081758",
                },
                {
                    "country": "EP",
                    "date": "20170927",
                    "kind": "A1",
                    "number": "3221665",
                },
            ],
            "status": [
                {
                    "code": "15",
                    "date": "20170825",
                    "description": "Request for examination was made",
                },
                {
                    "code": "17",
                    "date": "20170512",
                    "description": "The international publication has been made",
                },
            ],
            "title": "INERTIAL CAROUSEL POSITIONING",
        }
        assert pub.procedural_steps == [{'code': 'RFEE',
            'date': '20171113',
            'description': 'Renewal fee payment - 03',
            'phase': 'undefined'},
            {'code': 'FFEE',
            'date': '20170511',
            'description': 'Payment of national basic fee',
            'phase': 'entry-regional-phase'},
            {'code': 'SFEE',
            'date': '20170511',
            'description': 'Payment of the fee for supplementary search',
            'phase': 'entry-regional-phase'},
            {'code': 'DEST',
            'date': '20170511',
            'description': 'Payment of the first designation fee',
            'phase': 'entry-regional-phase'},
            {'code': 'EXAM_PCT',
            'date': '20170511',
            'description': 'Payment of the fee for examination',
            'phase': 'entry-regional-phase'},
            {'code': 'EXAM52',
            'date': '20170511',
            'description': 'Date on which the examining division has become responsible',
            'phase': 'examination'},
            {'code': 'ABEX',
            'date': '20171229',
            'description': 'Amendments - (claims and/or description)',
            'phase': 'examination'},
            {'code': 'PREX',
            'date': '20160819',
            'description': 'Preliminary examination - PCT II',
            'phase': 'international-examination'},
            {'code': 'PROL',
            'date': None,
            'description': 'Language of the procedure - en',
            'phase': 'examination'},
            {'code': 'ISAT',
            'date': None,
            'description': 'International searching authority - US',
            'phase': 'entry-regional-phase'}]




import re

SERIAL_RE = re.compile(r"[\d\-\/\,]{2,}")
NUMBER_CLEAN_RE = re.compile(r"[^\d]+")
PUNCTUATION_AND_WHITESPACE_CLEAN_RE = re.compile("\W+")
COUNTRY_CODE_RE = re.compile(r"^[A-Z]{2,3}")
KIND_CODE_RE = re.compile(r"[A-Z]\d?$")
COUNTRY_CODES = ["US", "EP", "PCT", "WO", 'CA']

"""
RESOURCES:

General Info: https://www.epo.org/searching-for-patents/helpful-resources.html
Numbering Systems: http://www.wipo.int/export/sites/www/standards/en/pdf/07-02-02.pdf
Kind Codes: http://www.wipo.int/export/sites/www/standards/en/pdf/07-03-02.pdf


"""

def parse(number, *args, **kwargs):
    if type(number) == str and "PCT" in number:
        return PCTApplication(number, *args, **kwargs)
    else:
        return PatentNumber(number, *args, **kwargs)

class PatentNumber:
    def __init__(self, number, country=None):
        self.number = None
        self.kind_code = None
        self.country = country
        self.number = number


        if type(number) == str:
            number = number.strip().upper()
            if "RE" in number:
                number = number.replace("US", "")
                self.number = PUNCTUATION_AND_WHITESPACE_CLEAN_RE.sub("", number)
                self._handle_us_number()
                return

            kind_code_match = KIND_CODE_RE.search(number)
            if kind_code_match:
                self.kind_code = kind_code_match.group()
            if not country:
                country_code_match = COUNTRY_CODE_RE.match(number)
                if country_code_match:
                    country_code = country_code_match.group()
                    self.country = country_code
                    number = SERIAL_RE.search(number).group()
                    self.number = PUNCTUATION_AND_WHITESPACE_CLEAN_RE.sub("", number)
            if self.country == 'US':
                self._handle_us_number()
            elif self.country == 'CA':
                self._handle_ca_number()
            else:
                self.number = NUMBER_CLEAN_RE.sub("", number)
                self._handle_us_number()

        if type(number) == int:
            self.number = str(number)
            self._handle_us_number()

    def __repr__(self):
        return f"<PatentNumber(type={self.type}, country={self.country}, number={self.number}, kind={self.kind_code})>"

    def _handle_us_number(self):
        if "RE" in self.number:
            self.country = "US"
            self.type = "patent"
            self.kind_code = "E1"
        elif int(self.number) > 90000000:
            self.country = "US"
            self.type = "pre-grant publication"
            self.kind_code = "A1"
        elif int(self.number) < 11000000:
            self.country = "US"
            self.type = "patent"
            self.kind_code = "B2"
        else:
            self.country = "US"
            self.type = "application"
            self.kind_code = ""
    
    def _handle_ca_number(self):
        if not self.kind_code:
            self.kind_code = 'A'

        ca_kind_codes = {
            'A': 'application',
            'A1': 'publication',
            'B': 'patent',
            'C': 'patent',
            'E': 'reissue patent',
            'F': 'reexamined patent',
        }
        self.type = ca_kind_codes[self.kind_code]

    def display(self):
        if self.country == "US":
            if self.type == "pre-grant publication":
                formatted_number = self.number[:4] + "/" + self.number[4:]
            elif self.type == "patent" and self.number.isdigit():
                formatted_number = "{:,}".format(int(self.number))
            elif self.type == "patent":
                formatted_number = self.number
            elif self.type == "application":
                formatted_number = (
                    self.number[:2] + "/" + "{:,}".format(int(self.number[2:]))
                )
            return f"US {formatted_number} {self.kind_code}".strip()
        elif self.country == "CA":
            return f'CA {self.number} {self.kind_code}'

    def abbreviation(self):
        if self.type == "pre-grant publication":
            return "U.S. Pub. Patent App."
        elif self.type == "patent":
            return "U.S. Patent"
        elif self.type == "application":
            return "U.S. Patent App."

    def __str__(self):
        return f"{self.country}{self.number}{self.kind_code}"


class PCTApplication:
    def __init__(self, number):
        chunks = number.split("/")
        if len(chunks) == 3:
            self.country, year = chunks[1][:2], chunks[1][2:]
            if len(year) == 2:
                if int(year) < 50:
                    year = "20" + year
                else:
                    year = "19" + year
            self.year = year
            self.number = chunks[2].rjust(6, "0")
            self.type = "international application"
            self.kind_code = ""

    def display(self, style="old"):
        if style == "old":
            return f"PCT/{self.country}{self.year[2:]}/{self.number[-5:]}"
        elif style == "new":
            return f"PCT/{self.country}{self.year}/{self.number}"
        else:
            raise ArgumentError()

    def __str__(self):
        return f"PCT{self.country}{self.year}{self.number}"

    def __repr__(self):
        return f"<PCTApplication(country_code={self.country}, year={self.year}, number={self.number})>"
